#!/usr/bin/env python
VERSION_TEXT = 'pypFFT R015 (C) 2013 Bernhard Slawik Digital'

# Major library imports
import pyaudio
from numpy import fromstring, short,	array,concatenate, hstack

#from ctypes import c_float	# Performance improvement by explicitly casting to C-types?

# Show code (exception) where values get converted (which is very slow)
from OpenGL.arrays import numpymodule
numpymodule.NumpyHandler.ERROR_ON_COPY = True

from scipy import fft

import math

import pygame
from pygame.locals import *

import OpenGL
OpenGL.ERROR_CHECKING = False	# Gives 3-4 fps
OpenGL.ERROR_LOGGING = False	# Enable for debugging
OpenGL.FULL_LOGGING = False	# Enable for debugging!
from OpenGL.GL import *
from OpenGL.GLU import *	# for gluPerspective
import gutil

from OpenGL.arrays import vbo	# VertexBufferObjects
#from OpenGL.arrays import *

lastPut = ''
def put(txt):
	print("pypFFT:	" + txt)
	global lastPut
	lastPut = txt

# Takes a text, returns a display list to draw it
textId = 1
def text(txt):
	col = (255,255,255,255)
	#col = (255,0,255,128)
	stringTex, w, h, tw, th = gutil.loadText(txt, font=FONT_NAME, size=FONT_SIZE, color=col, antialias=True)
	return [gutil.createTexDL(stringTex, tw, th), tw, th]


global SCREEN_SIZE
SCREEN_SIZE = (1024, 768)

SAMPLE_RATE = 2*48000	#11025
CHANNELS = 2

global NUM_BANDS, MAX_BANDS
MAX_BANDS = 1024*8	# maximum number of bands
NUM_BANDS = 512	# Default number of bands

global NUM_SAMPLES2, MAX_SAMPLES2
NUM_SAMPLES = 256	#NUM_BANDS * 2	# Audio buffer size (typically num_bands*2, but we have an intelligent ring-buffer :-D )
NUM_SAMPLES2 = NUM_BANDS * 2	# virtual buffer for FFT (always 2*bands)
MAX_SAMPLES2 = MAX_BANDS * 2	# virtual buffer for FFT (always 2*bands)
BUFFER_MAX = (NUM_SAMPLES*2*CHANNELS)*2	# In order to prevent input buffer overflows

NUM_STEREOSCAPE = 128	# Number of bars in stereoscape

global MONO
MONO = True if CHANNELS == 1 else False

# Time zooming allows for higher frequency resolution in lower bands. Since FFT is more detailed towards the higher frequencies we can use a ring buffer and perform the FFT on a time-stretched version of the wave form.
# The loss in resolution on the time domain is partially compensated by "fading out" the older samples (see scaleCache[])
global TIME_ZOOM
TIME_ZOOM = 3	# Oversampling (time stretch). Recommended is "2", use "1" for high-frequency analysis, high values for low-frequency analysis
MAX_TIME_ZOOM = 16
RING_SIZE = MAX_SAMPLES2 * MAX_TIME_ZOOM	# Size of ring buffer

global SHOW_SPECTRUM
SHOW_SPECTRUM = True
global SHOW_SPECTRUM_GRID, SHOW_SPECTRUM_COLOR
SHOW_SPECTRUM_GRID = True
SPECTRUM_GRID_STEPS_FRQ = 10	#18
SPECTRUM_GRID_STEPS_VOL = 14
SHOW_SPECTRUM_COLOR = True

global SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2
SHOW_LINES_MAX = True
SHOW_LINES_AVG = True
SHOW_LINES_AVG2 = True

global SHOW_SIDE_BY_SIDE
SHOW_SIDE_BY_SIDE = True

global SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2, SBS_SHOW_LINES_DIFF
SBS_SHOW_LINES_MAX = not True
SBS_SHOW_LINES_AVG = True
SBS_SHOW_LINES_AVG2 = not True
SBS_SHOW_LINES_DIFF = not True

global SHOW_HELP
SHOW_HELP = not True

global SHOW_INFO
SHOW_INFO = True

global SHOW_FPS
SHOW_FPS = True

GRID_COLOR = (0.33, 0.33, 0.33, 1.0)
SIGNAL_COLOR = (1.0, 1.0, 1.0, 1.0)
SIGNAL_COLOR_ELIMINATION = (1.0, 0.0, 0.0, 0.5)

INDICATOR_SIZE = 0.025
indicatorBright = 0.5
INDICATOR_COLOR = [SIGNAL_COLOR[i]*indicatorBright for i in xrange(4)]	#(1.0, 1.0, 1.0, 1.0)	#(0.625, 0.9, 0.625, 1.0)	#(0.75, 1.00, 0.75, 1.0)
INDICATOR_COLOR_ELIMINATION = [SIGNAL_COLOR_ELIMINATION[i]*indicatorBright for i in xrange(4)]	#(1.0, 0.0, 0.0, 1.0)	#(0.625, 0.9, 0.625, 1.0)	#(0.75, 1.00, 0.75, 1.0)




FONT_NAME = None
FONT_SIZE = 24	# Size for generating the font bitmaps
FONT_SCALE = 0.0015	# Scale in OpenGL
FONT_TRACK = 0.75	# tracking/kerning
SPACE = 0.05	# Space around border (for text)
SPECTRUM_BAR_GAP = 0.002	# Space between bars of spectrum

global PHON_CORRECTION, PHON_DELIN
PHON_CORRECTION = True	# Boost higher frequencies
PHON_DELIN = NUM_BANDS / 8	# For phon correction (high boost) - Pivot band. @TODO: Lock to FREQUENCY, not to BAND!!!
PHON_BASE = 3.0	# Strength of phon distortion (2-4)

global PHASE
PHASE = True	# Do phase analysis (phase correlation and stereoscape)
PHASE_SAMPLES = NUM_SAMPLES

LOG_FREQUENCY = 6.0	# Logarithmicy :-) of frequency scale (4 - 8.0)
LOG_FREQUENCY_SQUARE = math.pow(2.0, LOG_FREQUENCY) - 1.0

LOG_VOLUME = 1.5	# Logarithmacy of volume scale
LOG_VOLUME_SQUARE = math.pow(2.0, LOG_VOLUME) - 1.0

# "AVG" smoothes band decay whilst keeping the maximum peak - accomodates for the psychoacoustic effect of "Time Masking"
AVG_KEEP_FACTOR = 0.7	# 0.8
AVG_ADD_FACTOR = 1.0 - AVG_KEEP_FACTOR
SPECTRUM_AVG_COLOR = (0.625, 0.9, 0.625, 1.0)	#INDICATOR_COLOR

# "AVG2" averages bands slowly across time and frequency domain - accomodates for the psychoacoustic effect of "Frequency Masking"
AVG2_KEEP_FACTOR = 63.0/64.0
AVG2_ADD_FACTOR = 1.0 - AVG2_KEEP_FACTOR
AVG2_MAIN_FACTOR = 0.85	# How much of the center frequency to add to average
AVG2_NEIGHBOR_FACTOR = (1.0 - AVG2_MAIN_FACTOR) / 2.0	# How much of neighboring 2 frequencies to add

# For peak line
PEAK_STAY = 200
PEAK_FALLOFF_FACTOR = 0.995
PEAK_MAIN_FACTOR = 0.995
PEAK_NEIGHBOR_FACTOR = (1.0 - PEAK_MAIN_FACTOR) / 2.0	# How much of neighboring 2 frequencies to add

SPECTRUM_AVG2_COLOR = (1.0, 0.75, 0.25, 1.0)
SPECTRUM_PEAK_COLOR = (0.375, 0.125, 0.125, 1.0)
SPECTRUM_DIFF_COLOR = (0.5, 0.75, 1.0, 1.0)

global SONO
SONO = True
SONO_STEPS = 64
SONO_STEPS_INTERNAL = 2
SONO_SKIP_FRQ = 1
SONO_RANGE = 6.0	# z range

global SHOW_CHANNEL_WAVES
SHOW_CHANNEL_WAVES = True
global SHOW_MASTER_WAVE
SHOW_MASTER_WAVE = True
WAVE_SAMPLES = NUM_SAMPLES*2 #NUM_SAMPLES2	# Number of samples to visualize, e.g. "NUM_SAMPLES" for audio buffer, "NUM_SAMPLES2" for FFT buffer
WAVE_STATIC_FLIPS = True	# Allow the "static osc" feature to flip the waveform. if False: Always try to start the Waveform in one direction

global SHOW_ELIMINATION
SHOW_ELIMINATION = True

VIEW_POS = [0.0, 0.0, 0.0]
WIGGLE = False

max_time = float(NUM_SAMPLES) / SAMPLE_RATE
max_freq = float(SAMPLE_RATE / 2)

streamIn = None
pa = pyaudio.PyAudio()

"""
# math.sin is faster than my lookup table...
sinTabPow = 0x1000
sinTabScale2 = 2.0*math.pi / sinTabPow
sinTabScale = float(sinTabPow) / (2.0*math.pi)
sinTab = [math.sin(i*sinTabScale2) for i in xrange(sinTabPow)]
def _sin(a):
	return sinTab[int(round(a*sinTabScale)) % sinTabPow]
sinTabCosShift = int(sinTabPow / 4)
def _cos(a):
	return sinTab[int(round(a*sinTabScale) + sinTabCosShift) % sinTabPow]
"""

def bandToX(band):
	#return round( (math.log( 1.0 + float(band) / float(NUM_BANDS), 2.0) * screenH))
	return (math.log(1.0 + (float(band) / float(NUM_BANDS)) * LOG_FREQUENCY_SQUARE, 2.0) / LOG_FREQUENCY)

def volToY_calc(vol):
#def volToY(vol):
	#if (vol < 0): return -volToY(-vol)
	#return (math.log( 1.0 + vol * LOG_VOLUME_SQUARE, 2.0) / LOG_VOLUME)/4
	#return (math.log( 1.0 + vol * LOG_VOLUME_SQUARE, 2.0) / LOG_VOLUME)
	return (math.log( 1.0 + vol * LOG_VOLUME_SQUARE, 2.0) / LOG_VOLUME)*0.2

volToYRange = 100.0
volToYPow = 1024*8
volToYScale = float(volToYPow / volToYRange)
volToYCache = []
for i in xrange(volToYPow): volToYCache.append(volToY_calc(float(i) * volToYRange/volToYPow))
def volToY(vol):
	if (vol < 0): return -volToY(-vol)
	i = int(vol * volToYScale)
	if (i < volToYPow):		return volToYCache[i]
	else:				return volToYCache[volToYPow-1]


def frqToStr(frq):
	if (frq < 1000):	return str(int(frq)) +' Hz'
	else:				return str(round(frq/1000.0, 1)) + ' kHz'



def volToCol_calc(vol):
	vol = vol/4.0
	if (vol == 0): return (0.0,0.0,0.0)
	else:
		if (vol < 0.25): return (0.0, 0.0, 0.25 + vol*2.0)	# Blue
		else:
			vol = (vol - 0.25)/2
			if (vol < 0.25): return (0.0, vol*2, 0.75)	# Cyan
			else:
				vol = (vol - 0.25)/2
				if (vol < 0.25): return (vol, 0.5 + vol, 0.75-vol)	# Yellow
				else:
					vol = (vol - 0.25)/2
					if (vol < 0.25): return (0.25 + vol*3, 0.75 + vol, 0.5)
					else:
						vol = (vol - 0.25)/2
						if (vol < 0.25): return (1.0, 1.0, 0.5 + vol*4)
						else:
							return (1.0, 1.0, 1.0)
volToColRange = 100.0
volToColPow = 1024
volToColScale = float(volToColPow / volToColRange)
volToColCache = []
for i in xrange(volToColPow): volToColCache.append(volToCol_calc(float(i) * volToColRange/volToColPow))
def volToCol(vol):
	i = int(vol * volToColScale)
	if (i < 0): return volToCol(-vol)
	
	if (i < volToColPow):		return volToColCache[i]
	else:				return volToColCache[volToColPow-1]
def volCol(vol):
	#return
	if (i < 0): glColor3f(*volToColCache[-i])
	i = int(vol * volToColScale)
	if (i < volToColPow):		glColor3f(*volToColCache[i])
	else:				glColor3f(*volToColCache[volToColPow-1])


def angleBetween(deltaX, deltaY):
	an = 0.0
	if (deltaX == 0): #deltaX = -0.01;
		if (deltaY < 0): an = math.pi*2	#// pi*2.0
		else: an = math.pi*1.5
		#an=3.1415*1.499;
	else:
		an = math.atan(deltaY/deltaX)

	if (deltaX>0):
		an = an + math.pi
	return an

class pypFFT():
	def __init__(self):
		pygame.init()
		#self.screen = pygame.display.set_mode((screenW, screenH), 0, 32)
		self.screen = pygame.display.set_mode(SCREEN_SIZE, HWSURFACE|OPENGL|DOUBLEBUF|RESIZABLE)	#|FULLSCREEN
		pygame.display.set_caption(VERSION_TEXT)
		
		self.resize(SCREEN_SIZE[0], SCREEN_SIZE[1])
		
		#screen.fill((0x00, 0x00, 0x00))
		self.clock = pygame.time.Clock()
		
		# Pre-render glyphs
		self.glyphs = [None] * 256
		for i in xrange(256):
			c = chr(i)
			if (i > 32) and (i < 160):
				self.glyphs[i] = text(c)
		"""
		if USE_HTKFFT:
			self.pow = 2**16
			self.esn = math.sqrt(NUM_BANDS*2)
			self.nt = [[0] * (NUM_BANDS)] * (NUM_BANDS*2)
			self.sinTab = [0] * NUM_BANDS
			self.cosTab = [0] * NUM_BANDS
			for j in xrange(NUM_BANDS):
				self.sinTab[j] = int(round(math.sin(2.0 * math.pi * ((1+j) / NUM_BANDS)) * self.pow))
				self.cosTab[j] = int(round(math.cos(2.0 * math.pi * ((1+j) / NUM_BANDS)) * self.pow))
				jh = (j+2.0)/2.0
				for i in xrange(NUM_BANDS*2):
					self.nt[i][j] = int(round(jh * (i+1))) % NUM_BANDS
		"""
		
		#pygame.mouse.set_visible(False)
		self.run()
		
	"""
	def fft(self, data):
		if USE_HTKFFT:
			return self.htkFFT(data)
		else:
			return abs(fft(data))[:NUM_BANDS]
	
	def htkFFT(self, data):
		r = [0 for i in xrange(NUM_BANDS)]
		p = self.pow
		e = self.esn
		st = self.sinTab
		ct = self.cosTab
		for j in xrange(NUM_BANDS):
			fX = 0
			fY = 0
			for i in xrange(2*NUM_BANDS):
				v = data[i]
				nti = self.nt[i][j]
				fX = fX + (v * st[nti])
				fY = fY + (v * ct[nti])
			r[j] = int(round(math.sqrt( (fX / p)**2 + (fY / p)**2 ) / e))
		#return np.array(r, np.int32)
		return r
	"""
	
	def resize(self, width, height):	# OpenGL resize
		global SCREEN_SIZE
		SCREEN_SIZE = (width, height)
		
		glMatrixMode(GL_MODELVIEW)
		glLoadIdentity()
		#glMultMatrixf(rot.inverse().asList())
		#glTranslatef(0.0, 0.0, 0.0)
		
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		
		glViewport(0, 0, width, height)
		glDepthRange(0.0, 1.0)	# Default = (0, 1)
		
		glClearColor(0.0, 0.0, 0.0, 0.0)
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
		
		
		### POSIT knows "focal length", but OpenGL needs "field of view". Let's convert...
		adjacent = 1.0	#self.focLength	# Optical distance
		opposite = 4.0	#float(SCREEN_SIZE[1]) / 2.0
		fovy =  2.0 * math.atan(opposite / adjacent)
		aspect = 1.0
		gluPerspective((180.0/math.pi) * fovy, aspect, 0.01, 10.0)
		glScalef(1.0*aspect, 1.0, -1.0)	# That does the magic -- took me hours to figure it out...
		
		glTranslatef(0,0,0.265)
		
		
		
		glMatrixMode(GL_MODELVIEW)
		# HUD shading: None at all
		glDisable(GL_LIGHTING)
		glDisable(GL_DEPTH_TEST)
		glDisable(GL_LINE_SMOOTH);
		#glEnable(GL_ALPHA_TEST);
		#glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); 
		#glEnable(GL_BLEND);
		
	
	def drawText(self, text, pos, col = (0xff, 0xff, 0xff), rot=0.0):
		
		space = 5.0	# Default space for not-printable chars
		
		glPushMatrix()
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
		glEnable(GL_BLEND)
		glEnable(GL_TEXTURE_2D)
		glTranslatef(pos[0], pos[1], 0.0)
		glScalef(FONT_SCALE,FONT_SCALE ,FONT_SCALE )
		glRotatef(rot, 0.0,0.0,1.0)
		#glColor4f(col[0]/255.0, col[1]/255.0, col[2]/255.0, 1.0)
		
		for i in xrange(len(text)):
			g = self.glyphs[ord(text[i])]
			if not g == None:
				glCallList(g[0])
				glTranslatef(g[1]*FONT_TRACK, 0.0, 0.0)
				#glTranslatef(space, 0.0, 0.0)
			else:
				glTranslatef(space*FONT_TRACK, 0.0, 0.0)
		
		glDisable(GL_TEXTURE_2D)
		glDisable(GL_BLEND)
		glBlendFunc(GL_ONE, GL_ZERO)
		glPopMatrix()
		
	
	def drawWave(self, data, l, col=SIGNAL_COLOR, indicatorColor=INDICATOR_COLOR):
		xs = 1.0 / l
		dc = 0.0
		vMin = 1.0
		vMax = -1.0
		clip = False
		pointData = array([0.0] * (2*l), 'f')
		#glBegin(GL_POINTS)
		#glPointSize(2)
		j = 0
		for i in xrange(l):	#len(data)):
			x = i*xs
			v = data[i]
			#glVertex2f(x, v)
			pointData[j] = x
			j = j + 1
			pointData[j] = v
			j = j + 1
			
			dc = dc + v
			vMax = max(v, vMax)
			vMin = min(v, vMin)
		#glEnd()
		if (vMin <= -1.0) or (vMax >= 1.0): clip = True
		
		glColor4f(*col)
		pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')
		pointVbo.bind()
		try:
			glEnableClientState(GL_VERTEX_ARRAY)
			glDisableClientState(GL_COLOR_ARRAY)
			#glEnableClientState(GL_COLOR_ARRAY)
			glVertexPointer(2, GL_FLOAT, 0, pointVbo )
			glDrawArrays(GL_POINTS, 0, l)
		finally:
			pointVbo.unbind()
			glDisableClientState(GL_VERTEX_ARRAY)
		
		
		if (clip):
			glColor4f(1.0, 0.0, 0.0, 0.5)
		else:
			glColor4f(*GRID_COLOR)
		
		# Grid
		glBegin(GL_LINES)
		glVertex2f(0.0, -1.0)
		glVertex2f(1.0, -1.0)
		
		glVertex2f(0.0, 0.0)
		glVertex2f(1.0, 0.0)
		
		glVertex2f(0.0, 1.0)
		glVertex2f(1.0, 1.0)
		
		# Indicators
		if (clip):
			glColor4f(1.0, 0.0, 0.0, 0.5)
		else:
			glColor4f(*indicatorColor)
		
		# DC
		dc = dc / l
		glVertex2f(0.0, dc)
		glVertex2f(0.0-INDICATOR_SIZE, dc)
		glEnd()
		
		# Peak
		#glColor3f(*volToCol(max(vMax, -vMin)))
		
		glBegin(GL_QUADS)
		glVertex2f(1.0+INDICATOR_SIZE, vMin)
		glVertex2f(1.0, vMin)
		glVertex2f(1.0, vMax)
		glVertex2f(1.0+INDICATOR_SIZE, vMax)
		glEnd()
		
		
	
	def drawBars(self, data, colData, xCache):
		"""
		# Naive
		glBegin(GL_QUADS)
		for i in xrange(NUM_BANDS):
			x = xCache[i]
			x2 = max(x+SPECTRUM_BAR_GAP, xCache[i+1] - SPECTRUM_BAR_GAP)
			
			y = volToY(data[i])
			
			volCol(colData[i])
			glVertex2f(x, y)
			glVertex2f(x2, y)
			glVertex2f(x2, 0.0)
			glVertex2f(x, 0.0)
		glEnd()
		"""
		
		# Using VBO
		compsVertex = 4*2	# 4 points per bar, 2 components (x/y) per vertex
		compsColor = 4*3	# 4 points per bar, 3 color-components per vertex
		#pointData  = array([0.0] * (NUM_BANDS*sizeData), 'f')
		#pointColor = array([0.0] * (NUM_BANDS*compsColor), 'f')
		pointData  = array([0.0] * (NUM_BANDS*(compsVertex + compsColor)), 'f')
		
		#col0 = volToCol(0.0)
		
		jVertex = 0
		jColor = NUM_BANDS*compsVertex
		for i in xrange(NUM_BANDS):
			x = xCache[i]
			x2 = max(x+SPECTRUM_BAR_GAP, xCache[i+1] - SPECTRUM_BAR_GAP)
			
			y = volToY(data[i])
			#volCol(colData[i])
			
			# Store vertex data
			#pointVertex[jData:jData+sizeData] = [x,y, x2,y, x2,0.0, x,0.0]
			pointData[jVertex:jVertex+compsVertex] = [x,y, x2,y, x2,0.0, x,0.0]
			jVertex = jVertex + compsVertex
			
			if (SHOW_SPECTRUM_COLOR): 
				# Store color data
				col = volToCol(colData[i])
				pointData[jColor:jColor+compsColor] = [col[0],col[1],col[2], col[0],col[1],col[2], col[0],col[1],col[2], col[0],col[1],col[2]]
				
				#col0 = volToCol(colData[i]/2)
				#pointData[jColor:jColor+compsColor] = [col[0],col[1],col[2],	col[0],col[1],col[2],	col0[0],col0[1],col0[2],	col0[0],col0[1],col0[2]]
				jColor = jColor + compsColor
		
		glColor4f(*SIGNAL_COLOR)
		#pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')	# without color
		#stack = hstack((pointVertex, pointColor))
		#stack = stack.astype('float32')
		#pointVbo = vbo.VBO(stack, usage='GL_STATIC_DRAW')	# with color
		#pointVbo = vbo.VBO(hstack((pointData, pointColor)), usage='GL_STATIC_DRAW')	# with color
		#pointVbo = vbo.VBO(concatenate([pointData, pointColor]).astype('float32'), usage='GL_STATIC_DRAW')	# with color
		#pointVbo = vbo.VBO(concatenate([pointData, pointColor]), usage='GL_STATIC_DRAW')	# with color
		pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')	# generic
		
		pointVbo.bind()
		try:
			glEnableClientState(GL_VERTEX_ARRAY)
			glVertexPointer(2, GL_FLOAT, 0, pointVbo)
			if (SHOW_SPECTRUM_COLOR):
				glEnableClientState(GL_COLOR_ARRAY)	# only with color
				glColorPointer(3, GL_FLOAT, 0, pointVbo + (NUM_BANDS*compsVertex)*4)	# last 4 = size of 32bit float
			glDrawArrays(GL_QUADS, 0, NUM_BANDS*4)
		finally:
			pointVbo.unbind()
			glDisableClientState(GL_VERTEX_ARRAY)
			if (SHOW_SPECTRUM_COLOR): glDisableClientState(GL_COLOR_ARRAY)	# only with color
		
	
	def drawLine(self, data, xCache, col):
		l = 1+NUM_BANDS
		pointData = array([0.0] * (2*l), 'f')
		#glBegin(GL_LINE_STRIP)
		glColor4f(*col)
		#glVertex2f(0.0, 0.0)
		j = 2	# first: 0.0, 0.0 (already initialized in array)
		for i in xrange(NUM_BANDS):
			#glVertex2f(((xCache[i] + xCache[i+1]) / 2.0), volToY(data[i]))
			pointData[j] = ((xCache[i] + xCache[i+1]) / 2.0)
			j = j + 1
			pointData[j] = volToY(data[i])
			j = j + 1
		#glEnd()
		
		pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')
		pointVbo.bind()
		try:
			glEnableClientState(GL_VERTEX_ARRAY)
			glDisableClientState(GL_COLOR_ARRAY)
			#glEnableClientState(GL_COLOR_ARRAY)
			glVertexPointer(2, GL_FLOAT, 0, pointVbo )
			glDrawArrays(GL_LINE_STRIP, 0, l)
		finally:
			pointVbo.unbind()
			glDisableClientState(GL_VERTEX_ARRAY)
		
	
	def drawPhase(self, phase, col=SIGNAL_COLOR, indicatorColor=INDICATOR_COLOR):
		l = len(phase)
		ulfx=0.0
		ulfy=0.0
		urfx=0.0
		urfy=0.0
		dlfx=0.0
		dlfy=0.0
		drfx=0.0
		drfy=0.0
		
		# Grid
		glBegin(GL_LINES)
		glColor4f(*GRID_COLOR)
		glVertex3f(0.0, -0.5, 0.0)
		glVertex3f(0.0,  0.5, 0.0)
		
		glVertex3f(-0.5,  0.0, 0.0)
		glVertex3f( 0.5,  0.0, 0.0)
		glEnd()
		
		pointData = array([0.0] * (2*l), 'f')
		j = 0
		
		#glBegin(GL_POINTS)
		for i in xrange(l):
			fx = phase[i][0]
			fy = phase[i][1]
			#glVertex2f(fx, fy)
			pointData[j:j+2] = [fx, fy]
			j = j + 2
			
			if fx<=0.0:    ##### L
				if fy<=0.0 :  ### U
					ulfx = min(ulfx, fx)
					ulfy = min(ulfy, fy)
				else:	#if fy>=0.0 :  ### D
					dlfx = min(dlfx, fx)
					dlfy = max(dlfy, fy)
			else:	#if fx>=0.0:    ##### R
				if fy<=0.0 :  ### U
					urfx = max(urfx, fx)
					urfy = min(urfy, fy)
				else:	#if fy>=0.0 :  ### D
					drfx = max(drfx, fx)
					drfy = max(drfy, fy)
		#glEnd()
		
		glColor4f(*col)
		pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')
		pointVbo.bind()
		try:
			glEnableClientState(GL_VERTEX_ARRAY)
			#glDisableClientState(GL_COLOR_ARRAY)
			#glEnableClientState(GL_COLOR_ARRAY)
			glVertexPointer(2, GL_FLOAT, 0, pointVbo )
			glDrawArrays(GL_POINTS, 0, l)
		finally:
			pointVbo.unbind()
			glDisableClientState(GL_VERTEX_ARRAY)
		
		
		glBegin(GL_QUADS);
		glColor4f(*indicatorColor)
		glVertex3f(ulfx,-0.5-INDICATOR_SIZE, 0.0)
		glVertex3f(urfx,-0.5-INDICATOR_SIZE, 0.0)
		glVertex3f(urfx,-0.5 , 0.0)
		glVertex3f(ulfx,-0.5 , 0.0)

		glVertex3f(dlfx, 0.5+INDICATOR_SIZE, 0.0)
		glVertex3f(drfx, 0.5+INDICATOR_SIZE, 0.0)
		glVertex3f(drfx, 0.5 , 0.0)
		glVertex3f(dlfx, 0.5 , 0.0)

		glVertex3f(-0.5-INDICATOR_SIZE, ulfy, 0.0)
		glVertex3f(-0.5 , ulfy, 0.0)
		glVertex3f(-0.5 , dlfy, 0.0)
		glVertex3f(-0.5-INDICATOR_SIZE, dlfy, 0.0)

		glVertex3f(0.5+INDICATOR_SIZE, urfy, 0.0)
		glVertex3f(0.5 , urfy, 0.0)
		glVertex3f(0.5 , drfy, 0.0)
		glVertex3f(0.5+INDICATOR_SIZE, drfy, 0.0)
		glEnd()
		
	
	def drawStereoscape(self, stereoscape):
		stereo = 0.0
		l = len(stereoscape)
		# Using VBO
		#pointData = array([0.0] * (NUM_STEREOSCAPE*4*2), 'f')	# monochrome
		compsVertex = 4*2	# 4 points per bar, 2 components (x/y) per vertex
		compsColor = 4*3	# 4 points per bar, 3 color-components per vertex
		pointData  = array([0.0] * (NUM_STEREOSCAPE*(compsVertex + compsColor)), 'f')
		
		jVertex = 0
		jColor = NUM_STEREOSCAPE*compsVertex
		xs = (1.0/NUM_STEREOSCAPE)
		for i in xrange(NUM_STEREOSCAPE):
			x = -0.5 + i*xs
			x2 = x+xs
			v = (stereoscape[i] / NUM_SAMPLES)
			y = volToY(v * NUM_BANDS * 16)
			#y = (v * NUM_BANDS * 8)
			
			#volCol(v*65536.0)
			
			#glVertex2f(x, -y)
			#glVertex2f(x2, -y)
			#glVertex2f(x2, 0.0)
			#glVertex2f(x, 0.0)
			
			# Store vertex data
			pointData[jVertex:jVertex+compsVertex] = [x,-y,	x2,-y,	x2,0.0,	x,0.0]
			jVertex = jVertex + compsVertex
			
			if (SHOW_SPECTRUM_COLOR): 
				# Store color data
				col = volToCol(v*65535)
				pointData[jColor:jColor+compsColor] = [col[0],col[1],col[2], col[0],col[1],col[2], col[0],col[1],col[2], col[0],col[1],col[2]]
				jColor = jColor + compsColor
			
			stereo = stereo + (x*v)
		#glEnd()
		
		# Draw signal using vbo
		glColor4f(*SIGNAL_COLOR)
		pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')
		pointVbo.bind()
		try:
			glEnableClientState(GL_VERTEX_ARRAY)
			glVertexPointer(2, GL_FLOAT, 0, pointVbo)
			if (SHOW_SPECTRUM_COLOR):
				glEnableClientState(GL_COLOR_ARRAY)
				glColorPointer(3, GL_FLOAT, 0, pointVbo + (NUM_STEREOSCAPE*compsVertex)*4)	# last 4 = size of 32bit float
			glDrawArrays(GL_QUADS, 0, NUM_STEREOSCAPE*4)
		finally:
			pointVbo.unbind()
			glDisableClientState(GL_VERTEX_ARRAY)
			if (SHOW_SPECTRUM_COLOR): glDisableClientState(GL_COLOR_ARRAY)
		
		
		stereo = stereo * 1024.0
		if stereo < 0	: stereo = -volToY(-stereo)
		else		: stereo = volToY(stereo)
		# Draw stereo indicator
		
		#glColor4f(*SIGNAL_COLOR)
		glBegin(GL_QUADS)
		glColor4f(*INDICATOR_COLOR)
		glVertex2f(stereo, SPACE)
		glVertex2f(0.0, SPACE)
		glVertex2f(0.0, 0.0)
		glVertex2f(stereo, 0.0)
		glEnd()
		
		
		# Grid on-top
		glColor4f(*GRID_COLOR)
		glBegin(GL_LINES)
		glVertex2f(0.0, -0.5)
		glVertex2f(0.0, 0.0)
		glEnd()
		
		
	def drawSonoRow(self, sonoRow, sonoRow2, xCache):
		
		bands = NUM_BANDS/SONO_SKIP_FRQ
		
		compsVertex = 2*3	# 2 points per bar, 3 components (x/y/y) per vertex
		compsColor = 2*3	# 2 points per bar, 3 color-components per vertex
		pointData  = array([0.0] * (bands*(compsVertex + compsColor)), 'f')
		glEnableClientState(GL_VERTEX_ARRAY)
		glEnableClientState(GL_COLOR_ARRAY)
		zStep = 1.0 / SONO_STEPS
		
		#j2 = j+1
		#z = j*zStep
		z = zStep
		
		jVertex = 0
		jColor = bands*compsVertex
		i2 = 0
		#for i in xrange(int(min(NUM_BANDS, len(sono[j]), len(sono[j2]))/SONO_SKIP_FRQ)): #(bands):
		m = int(min(NUM_BANDS, len(sonoRow), len(sonoRow2))/SONO_SKIP_FRQ)
		while (i2 < m):
			#i2 = i * SONO_SKIP_FRQ
			
			x = xCache[i2]
			#x = i2/NUM_BANDS
			
			v = sonoRow[i2]
			y = volToY(v)
			
			v2 = sonoRow2[i2]
			y2 = volToY(v2)
			
			#volCol(v)
			#glVertex3f(x, y, z-zStep)
			
			pointData[jVertex:jVertex+compsVertex] = [x, y, 0.0, x, y2, zStep]
			#pointData[jVertex:jVertex+compsVertex] = [x, y2, zStep, x, y, 0.0]
			jVertex = jVertex + compsVertex
			
			# Store color data
			col = volToCol(v)
			col2 = volToCol(v2)
			#pointData[jColor:jColor+compsColor] = [col[0],col[1],col[2], col2[0],col2[1],col2[2]]
			pointData[jColor:jColor+compsColor] = [col[1]/2,col[0]/2,col[2]/2, col2[1]/2,col2[0]/2,col2[2]/2]
			jColor = jColor + compsColor
			
			i2 = i2 + SONO_SKIP_FRQ
		#glEnd()
		# Draw signal using vbo
		pointVbo = vbo.VBO(pointData, usage='GL_STATIC_DRAW')
		pointVbo.bind()
		glVertexPointer(3, GL_FLOAT, 0, pointVbo)
		glColorPointer(3, GL_FLOAT, 0, pointVbo + (bands*compsVertex)*4)	# last 4 = size of 32bit float
		glDrawArrays(GL_QUAD_STRIP, 0, bands)
		
		pointVbo.unbind()
		glDisableClientState(GL_VERTEX_ARRAY)
		glDisableClientState(GL_COLOR_ARRAY)
		#
	########################################################################
	########################################################################
	########################################################################
	
	def run(self):
		#screen = self.screen
		global NUM_BANDS, MAX_BANDS, MONO, TIME_ZOOM, PHASE_SAMPLES, NUM_SAMPLES2, MAX_SAMPLES2, RING_SIZE, PHON_CORRECTION, PHON_DELIN
		
		#ringData = [[0.0] * RING_SIZE, [0.0] * RING_SIZE]	##ringData = [[0.0] * RING_SIZE] * CHANNELS
		#ringData = [[0.0] * RING_SIZE] * CHANNELS
		ringData = [[], []]
		for i in xrange(RING_SIZE):
			ringData[0].append(0.0)
			ringData[1].append(0.0)
		
		ringO = 0
		data = [[0.0] * MAX_SAMPLES2, [0.0] * MAX_SAMPLES2]	#[[0.0] * NUM_SAMPLES2] * CHANNELS
		dataMono = [0.0] * WAVE_SAMPLES
		dataEliminated = [0.0] * WAVE_SAMPLES
		
		spectrum = [[0.0] * MAX_BANDS, [0.0] * MAX_BANDS]	#[[0.0] * NUM_BANDS] * CHANNELS
		#phase = [[0.0, 0.0] for i in xrange(NUM_SAMPLES)]
		phase = [[0.0, 0.0] for i in xrange(PHASE_SAMPLES)]
		phaseEliminated = [[0.0, 0.0] for i in xrange(PHASE_SAMPLES)]
		stereoscape = [0.0] * MAX_BANDS	# Cummulated
		#stereoscapePeak = [0.0] * MAX_BANDS	# Peak
		peakData = [[0.0] * MAX_BANDS, [0.0] * MAX_BANDS]	#peakData = [[0.0] * NUM_BANDS] * CHANNELS
		peakStay = [[0] * MAX_BANDS, [0] * MAX_BANDS]		#peakStay = [[0.0] * NUM_BANDS] * CHANNELS
		avgData = [[0.0] * MAX_BANDS, [0.0] * MAX_BANDS]	#[[0.0] * NUM_BANDS] * CHANNELS
		avg2Data = [[0.0] * MAX_BANDS, [0.0] * MAX_BANDS]	#[[0.0] * NUM_BANDS] * CHANNELS
		diffData = [0.0] * MAX_BANDS	#[[0.0] * NUM_BANDS]
		#sono = [[0.0] * MAX_BANDS] * SONO_STEPS	# If storing the whole sonogram is what you want. Else: use glDisplayLists!
		sono = [[0.0] * MAX_BANDS] * SONO_STEPS_INTERNAL
		sonoList = [glGenLists(1) for i in xrange(SONO_STEPS)]
		sonoListO = 0
		
		"""
		
		Following code was moved inside the loop ("if reCache")
		 
		# Generate mapping of band -> x-coordinate
		bandToXCache = [bandToX(i) for i in xrange(NUM_BANDS+1)]
		bandToXCacheLin = [float(i)/NUM_BANDS for i in xrange(NUM_BANDS+1)]
		phonCache = [(256.0/NUM_BANDS) * (PHON_BASE**((float(i)-PHON_DELIN) / (NUM_BANDS))) for i in xrange(NUM_BANDS)]
		
		# Generate mapping of sample -> volume (for time zoom ring factor)
		#data.append( ringData[j % RING_SIZE] )	# Don't scale (may look like weird dynamics)
		#data.append( ringData[j % RING_SIZE] * (float(i) / NUM_SAMPLES)	# scale everything
		#data.append( ringData[j % RING_SIZE] * (float(i) / NUM_SAMPLES2) )	# Scale towards present
		#data.append( ringData[j % RING_SIZE] * min(1.0, (float(i) / (NUM_SAMPLES2 - NUM_SAMPLES/TIME_ZOOM)) ))	# Scale all "old" frames
		if (NUM_SAMPLES2 <= NUM_SAMPLES/TIME_ZOOM):
			scaleCache = [1.0] * NUM_SAMPLES2
		else:
			scaleCache = [(min(1.0, (float(i) / (NUM_SAMPLES2 - NUM_SAMPLES/TIME_ZOOM)) ))	for i in xrange(NUM_SAMPLES2)]	# Scale all "old" frames]
		"""
		
		# Pre-render help
		glPushMatrix()
		helpList = glGenLists(1)
		glNewList(helpList, GL_COMPILE)
		
		
		x = -0.5
		y = 0.5
		s = SPACE/2
		glScalef(2.0, 2.0, 1.0)
		glColor3f(1.0,1.0,1.0)
		self.drawText(VERSION_TEXT, (x, y - 0*s))
		self.drawText("0-9 = presets", (x, y - 2*s))
		self.drawText("* / = time zoom", (x, y - 3*s))
		self.drawText("a = main analyzer", (x, y - 4*s))
		self.drawText("b = side-by-side", (x, y - 5*s))
		self.drawText("c = colors", (x, y - 6*s))
		self.drawText("d = stereo diff", (x, y - 7*s))
		self.drawText("e = phase elimination", (x, y - 8*s))
		self.drawText("f = fps", (x, y - 9*s))
		self.drawText("g = grid", (x, y - 10*s))
		self.drawText("h = help", (x, y - 11*s))
		self.drawText("i = info", (x, y - 12*s))
		self.drawText("l = lines (shift for side-by-side)", (x, y - 13*s))
		self.drawText("m = force mono", (x, y - 14*s))
		self.drawText("p = phon correction (boost HF)", (x, y - 15*s))
		self.drawText("s = sonogram (water fall)", (x, y - 16*s))
		self.drawText("v = phase correlation", (x, y - 17*s))
		self.drawText("w = channel waves", (x, y - 18*s))
		glEndList()
		glPopMatrix()
		
		gridList = None	# used for caching the grid as glDisplayList
		infoList = None
		reCache = True
		
		
		
		put("Opening audio... (CHANNELS=" + str(CHANNELS) + ", SAMPLE_RATE=" + str(SAMPLE_RATE) + ", NUM_SAMPLES=" + str(NUM_SAMPLES) + ")")
		streamIn = pa.open(format=pyaudio.paInt16, channels=CHANNELS, rate=SAMPLE_RATE, input=True, frames_per_buffer=NUM_SAMPLES)
		glPushMatrix()
		
		running = True
		frame = 0.0
		put("Running...")
		while running:
			
			if reCache:
				# Re-cache
				NUM_SAMPLES2 = NUM_BANDS * 2
				#RING_SIZE = NUM_SAMPLES2 * MAX_TIME_ZOOM	# Size of ring buffer
				PHON_DELIN = NUM_BANDS / 8
				bandToXCache = [bandToX(i) for i in xrange(NUM_BANDS+1)]
				bandToXCacheLin = [float(i)/NUM_BANDS for i in xrange(NUM_BANDS+1)]
				phonCache = [ (PHON_BASE**(PHON_BASE*(float(i)-PHON_DELIN) / (NUM_BANDS*TIME_ZOOM))) for i in xrange(NUM_BANDS+1)]
				if (NUM_SAMPLES2 <= NUM_SAMPLES/TIME_ZOOM):
					scaleCache = [1.0] * NUM_SAMPLES2
				else:
					scaleCache = [(min(1.0, (float(i) / (NUM_SAMPLES2 - NUM_SAMPLES/TIME_ZOOM)) ))	for i in xrange(NUM_SAMPLES2)]	# Scale all "old" frames]
				#NUM_BANDS_SQRT = math.sqrt(NUM_BANDS)
				gridList = None
				infoList = None
				
				reCache = False
			
			
			#pygame.time.delay(1000*NUM_SAMPLES/SAMPLE_RATE)
			tick = self.clock.tick()
			frame = frame + tick/1000.0
			
			
			# Clear the screen
			#screen.fill((0x00, 0x00, 0x00))	# 2D
			glClearColor(0.0, 0.0, 0.0, 1.0)	# OpenGL
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			
			glPopMatrix()
			glPushMatrix()
			glTranslatef(*VIEW_POS)
			glPushMatrix()
			
			# Wiggle...
			if WIGGLE:
				a = math.sin(frame/2.0)
				glTranslate(a * 0.2, 0.0, 0.0)
				glRotatef(a * 1.0, 0.0, 1.0, 0.0)
			
			
			# Get audio data
			#dataRaw = get_audio_data()
			
			waiting = streamIn.get_read_available()
			"""
			if (waiting > BUFFER_MAX):	#16*BUFFER_SIZE*CHANNELS*SAMPLE_WIDTH):
				put("Skipping old buffer (" + str(waiting) + " bytes)")
				streamIn.read(waiting - BUFFER_MAX)
			"""
			
			
			raw_data = streamIn.read(waiting)
			audio_data  = fromstring(raw_data, dtype=short)
			"""
			valid = False
			while not valid:
				try:
					#raw_data = streamIn.read(NUM_SAMPLES)
					raw_data = streamIn.read(waiting)
					audio_data  = fromstring(raw_data, dtype=short)
					valid = True
				except:
					put("Audio buffer overflow")
					valid = False
			
			audio_data  = fromstring(raw_data, dtype=short)
			"""
			
			if CHANNELS == 2:
				dataRaw = [audio_data[0::2] / 32768.0, audio_data[1::2] / 32768.0]
			else:
				dataRaw = [audio_data / 32768.0]
			
			
			############ Calculate... ###############
			for chn in xrange(CHANNELS if not MONO else 1):
				# Update ring buffer
				samples = len(dataRaw[chn])
				#ringData[ringO:ringO+NUM_SAMPLES-1] = data[:]
				#ringData[chn][ringO:ringO + NUM_SAMPLES] = [dataRaw[chn][i] for i in xrange(NUM_SAMPLES)]
				
				for i in xrange(samples):
					ringData[chn][(ringO + i) % RING_SIZE] = dataRaw[chn][i]
				
				
				# Generate (time zoomed) wave data from ring buffer
				data[chn] = [0.0] * NUM_SAMPLES2	#dataRaw[:]
				j = ringO + samples - (NUM_SAMPLES2 * TIME_ZOOM) + RING_SIZE*4
				for i in xrange(NUM_SAMPLES2):
					#data[chn].append( ringData[chn][j % RING_SIZE] * scaleCache[i])
					data[chn][i] = ringData[chn][j % RING_SIZE] * scaleCache[i]
					j = j + TIME_ZOOM
				
				# Calculate FFT
				spectrum[chn] = abs(fft(data[chn]))[:NUM_BANDS]
				#spectrum[chn] = self.fft(data[chn])
				
				#global PHON_CORRECTION
				if PHON_CORRECTION:
					# Phon-correction (i.e. boost high frequencies)
					#@FIXME: This must be fixed to a FREQUENCY (psychoacoutic!), not to a band number!
					#for i in xrange(NUM_BANDS): spectrum[chn][i] = spectrum[chn][i] * (i+PHON_DELIN) / (NUM_BANDS)
					#for i in xrange(NUM_BANDS): spectrum[chn][i] = spectrum[chn][i] * (PHON_BASE**((float(i)-PHON_DELIN) / (NUM_BANDS)))
					#for i in xrange(NUM_BANDS): spectrum[chn][i] = spectrum[chn][i] * (PHON_BASE**((float(i)-PHON_DELIN) / (NUM_BANDS)))
					for i in xrange(NUM_BANDS): spectrum[chn][i] = spectrum[chn][i] * phonCache[i]
				
				# Calculate band Peak
				#peakData[chn] = [max(spectrum[chn][i], peakData[chn][i] * PEAK_FALL_OFF_FACTOR) for i in xrange(NUM_BANDS)]
				norm = 16.0 / math.sqrt(NUM_BANDS)
				for i in xrange(NUM_BANDS):
					spectrum[chn][i] = spectrum[chn][i] * norm	# Normalize
					m = peakData[chn][i]
					
					if (spectrum[chn][i] >= m):
						peakData[chn][i] = spectrum[chn][i]
						peakStay[chn][i] = PEAK_STAY
						
					else:
						if (peakStay[chn][i] <= 0):
							peakData[chn][i]= max(peakData[chn][i] * PEAK_FALLOFF_FACTOR, \
							(\
								  PEAK_NEIGHBOR_FACTOR * (0 if (i<1) else peakData[chn][i-1])\
								+ PEAK_MAIN_FACTOR * peakData[chn][i] * PEAK_FALLOFF_FACTOR\
								+ PEAK_NEIGHBOR_FACTOR * (0 if (i>=NUM_BANDS-1) else peakData[chn][i+1])
							))
						else:
							peakStay[chn][i] = peakStay[chn][i] - 1
				
				# Calculate band Avg
				#avgData[chn] = [max(spectrum[chn][i], avgData[chn][i] * AVG_KEEP_FACTOR + spectrum[chn][i] * AVG_ADD_FACTOR) for i in xrange(NUM_BANDS)]
				avgData[chn][:NUM_BANDS] = [max(spectrum[chn][i], avgData[chn][i] * AVG_KEEP_FACTOR + spectrum[chn][i] * AVG_ADD_FACTOR) for i in xrange(NUM_BANDS)]
				
				# Calculate band Avg2 (slow avg)
				#avg2Data = [(avg2Data[i] * AVG2_KEEP_FACTOR + spectrum[i] * AVG2_ADD_FACTOR) for i in xrange(NUM_BANDS)]
				for i in xrange(NUM_BANDS):
					avg2Data[chn][i] =\
						(\
							  AVG2_NEIGHBOR_FACTOR * (0 if (i<1) else avg2Data[chn][i-1])\
							+ AVG2_MAIN_FACTOR  * avg2Data[chn][i]\
							+ AVG2_NEIGHBOR_FACTOR * (0 if (i>=NUM_BANDS-1) else avg2Data[chn][i+1])\
						) * AVG2_KEEP_FACTOR + spectrum[chn][i] * AVG2_ADD_FACTOR
			
			# Back to non-channel-iteration mode (i.e.: now perform operations on both channels at once)
			global SBS_SHOW_LINES_DIFF
			if SBS_SHOW_LINES_DIFF:
				if (CHANNELS > 1) and (not MONO):
					diffData = [(spectrum[1][i] - spectrum[0][i]) for i in xrange(NUM_BANDS)]
			
			
			ringO = (ringO + samples) % RING_SIZE
			
			# Sonogram
			global SONO
			if (SONO):
				for i in xrange(SONO_STEPS_INTERNAL-1):
					j = (SONO_STEPS_INTERNAL-1) - i
					sono[j] = sono[j-1]
				
				if MONO:	#CHANNELS == 1:
					sono[0] = [avgData[0][i] for i in xrange(NUM_BANDS)]
					#sono[0][0:NUM_BANDS] = [avgData[0][i] for i in xrange(NUM_BANDS)]
				else:
					sono[0] = [(avgData[0][i]+avgData[1][i])/2 for i in xrange(NUM_BANDS)]
					#sono[0][0:NUM_BANDS] = [(avgData[0][i]+avgData[1][i])/2 for i in xrange(NUM_BANDS)]
			
			# Calculate phase(s) and stereoscape
			global PHASE
			if (PHASE) and (CHANNELS > 1) and (not MONO):
				radScale = NUM_STEREOSCAPE / math.pi
				for i in xrange(NUM_STEREOSCAPE):
					stereoscape[i] = stereoscape[i] * AVG_KEEP_FACTOR
					#stereoscapePeak[i] = stereoscapePeak[i] * AVG_KEEP_FACTOR
				#i2 = (ringO - NUM_SAMPLES) + RING_SIZE
				i2 = (ringO + RING_SIZE - PHASE_SAMPLES)
				#for i in xrange(NUM_SAMPLES):
				for i in xrange(PHASE_SAMPLES):
					# Normal
					# Use FFT wave
					#x  = (data[0][i] * 128.0)
					#x2 = (data[1][i] * 128.0)
					
					# Use ring buffered original wave
					o = (i2 + i) % RING_SIZE
					x  = (ringData[0][o] * 128.0)
					x2 = (ringData[1][o] * 128.0)
					
					z  = (x2     - x     ) / 1024.0
					z1 = (abs(x) + abs(x2)) / 256.0
					z2 = (x      + x2     ) / 1024.0
					if not (z1 == 0.0):
						z = z/z1
						
						fx = math.sin(z) * z1 * 2.0
						fy = math.cos(z) * z2 * 2.0
						
						#phase[i] = [fx, fy]
						phase[i][0] = fx
						phase[i][1] = fy
						
						
						### stereoscape
						z = angleBetween(fx, fy)
						#j = (int(round(-z * radScale)) + (NUM_BANDS << 4) ) % NUM_STEREOSCAPE
						j = int(round(-z * radScale)) % NUM_STEREOSCAPE
						#j = (int(round((-z * radScale) / 2)) ) % NUM_STEREOSCAPE
						#if (j>=bands) then showmessage('err j<0 = '+inttostr(j));
						#k = (math.sqrt(fx*fx)+math.sqrt(fy*fy)) / NUM_STEREOSCAPE
						k = (math.sqrt(fx*fx + fy*fy)) / NUM_STEREOSCAPE
						#stereoscape[j] = max(k, stereoscape[j] * AVG_KEEP_FACTOR)
						stereoscape[j] = stereoscape[j] + k
						
						#if (k > stereoscapePeak[j]): stereoscapePeak[j] = k
					
					# Phase elimination
					z  = (x2     + x     ) / 1024.0
					z2 = (x      - x2    ) / 1024.0
					z1 = z2 * 4
					
					if not (z1 == 0.0):
						z = -(z/z1)
						fy = math.sin(z) * z1 *2.0
						fx = math.cos(z) * z2 *2.0
						
						#phaseEliminated[i]= [fx, fy]
						phaseEliminated[i][0] = fx
						phaseEliminated[i][1] = fy
			
			
			
			# Find zero pass
			imax = int(WAVE_SAMPLES /2)	# maximum look-back
			ringSyncO = (ringO - WAVE_SAMPLES*TIME_ZOOM - imax) + RING_SIZE
			i = 0
			thresh = 0.1
			chn = 0
			if (WAVE_STATIC_FLIPS) or (ringData[chn][(ringSyncO+i) % RING_SIZE] > 0):
				while (i < imax) and (ringData[chn][(ringSyncO+i) % RING_SIZE] <= thresh): i = i + 1
				#while (i < imax) and (ringData[chn][(j+i) % RING_SIZE] > -thresh): i = i + 1	# Optional
				while (i < imax) and (ringData[chn][(ringSyncO+i) % RING_SIZE] > 0): i = i + 1
			else:
				while (i < imax) and (ringData[chn][(ringSyncO+i) % RING_SIZE] >= -thresh): i = i + 1
				#while (i < imax) and (ringData[chn][(j+i) % RING_SIZE] > -thresh): i = i + 1	# Optional
				while (i < imax) and (ringData[chn][(ringSyncO+i) % RING_SIZE] < 0): i = i + 1
			if (i >= imax): i = 0
			ringSyncO = ringSyncO + i
			
			
			
			###################### Draw ################
			width = 2.0/(CHANNELS if not MONO else 1)
			height = 2.0/3
			yStart = -1.0 + height
			
			
			
			# Draw sono
			#global SONO
			if SONO:
				glPushMatrix()
				glLoadIdentity()
				#glTranslatef(0.0, -0.5, 0.0)
				glTranslatef(0.0, -1.0, 0.0)
				
				glRotatef(-75.0, 1.0,0.0,0.0)
				
				glTranslatef(-0.75, 0.0, SONO_RANGE)	# + 0.25)
				glScalef(2.0, 0.25, SONO_RANGE)
				
				sonoListO = (sonoListO + 1) % SONO_STEPS
				sonoListNew = sonoList[sonoListO]
				glNewList(sonoListNew, GL_COMPILE)
				self.drawSonoRow(sonoRow=sono[0], sonoRow2=sono[1], xCache=bandToXCache)
				#self.drawSonoRow(sonoRow=sono[1], sonoRow2=sono[0], xCache=bandToXCache)
				glEndList()
				zStep = 1.0 / SONO_STEPS
				"""
				for j in xrange(SONO_STEPS-1):
					glTranslatef(0.0, 0.0, zStep)
					glCallList(sonoList[(sonoListO-j+SONO_STEPS) % SONO_STEPS])
				"""
				for j in xrange(SONO_STEPS-1):
					glTranslatef(0.0, 0.0, -zStep)
					glCallList(sonoList[(sonoListO+j+1) % SONO_STEPS])
				
				glPopMatrix()
			
			#global PHASE
			if PHASE and (not MONO):
				# Draw stereoscape
				"""
				glPushMatrix()
				glTranslatef(0.0, 1.0, 0.0)
				glScalef(0.75, 1.0, 1.0)
				self.drawStereoscape(stereoscapePeak)
				glPopMatrix()
				"""
				
				glPushMatrix()
				glTranslatef(0.0, 1.0, 0.0)
				glScalef(0.75, 0.5, 1.0)
				self.drawStereoscape(stereoscape)
				glPopMatrix()
			
			global SHOW_ELIMINATION
			global SHOW_MASTER_WAVE
			if SHOW_MASTER_WAVE:
				# Draw mono/eliminated wave
				glPushMatrix()
				glTranslatef(0.5, 0.75, 0.0)
				glScalef(0.5, 0.25, 1.0)
				#glRotatef(-90, 0,0,1)
				
				if (CHANNELS == 1) or MONO:
					#dataMono = [ringData[0][(ringSyncO+k*TIME_ZOOM) % RING_SIZE] for k in xrange(WAVE_SAMPLES)]
					dataMono = [ringData[0][(ringO - WAVE_SAMPLES+k) % RING_SIZE] for k in xrange(WAVE_SAMPLES)]
					
					glColor4f(*SIGNAL_COLOR)
					self.drawWave(dataMono, WAVE_SAMPLES)
				else:
					for i in xrange(WAVE_SAMPLES):
						o = (ringSyncO+i*TIME_ZOOM) % RING_SIZE
						v1 = ringData[0][o]
						v2 = ringData[1][o]
						dataMono[i] = (v1+v2)/2
						dataEliminated[i] = (v1-v2)/2
					
					#glColor4f(*SIGNAL_COLOR)
					self.drawWave(dataMono, WAVE_SAMPLES, col=SIGNAL_COLOR, indicatorColor=INDICATOR_COLOR)
					
					if SHOW_ELIMINATION:
						#glColor4f(*SIGNAL_COLOR_ELIMINATION)
						self.drawWave(dataEliminated, WAVE_SAMPLES, col=SIGNAL_COLOR_ELIMINATION, indicatorColor=INDICATOR_COLOR_ELIMINATION)
				glPopMatrix()
			
			#global PHASE
			if (PHASE) and (CHANNELS > 1) and (not MONO):
				# Draw phase graph
				glPushMatrix()
				glTranslatef(-0.75, 0.75, 0.0)
				glScalef(0.5, 0.5, 1.0)
				
				#glColor4f(*SIGNAL_COLOR)
				self.drawPhase(phase, col=SIGNAL_COLOR, indicatorColor=INDICATOR_COLOR)
				
				if SHOW_ELIMINATION:
					glColor4f(*SIGNAL_COLOR_ELIMINATION)
					self.drawPhase(phaseEliminated, col=SIGNAL_COLOR_ELIMINATION, indicatorColor=INDICATOR_COLOR_ELIMINATION)
				glPopMatrix()
			
			
			#chn = 0
			for chn in xrange(CHANNELS if not MONO else 1):
				xStart = -1.0 + width*chn
				
				
				global SHOW_CHANNEL_WAVES
				if SHOW_CHANNEL_WAVES:
					# Draw channel wave
					glPushMatrix()
					glTranslatef(-1.0 + 2*chn, 0.0, 0.0)
					glScalef(0.5 * (1-2*chn), height/4, 1.0)
					glColor3f(1.0, 1.0, 1.0)
					self.drawWave([ringData[chn][(ringSyncO+k*TIME_ZOOM) % RING_SIZE] for k in xrange(WAVE_SAMPLES)], WAVE_SAMPLES)
					glPopMatrix()
					
					"""
					# Raw wave
					glPushMatrix()
					glColor3f(1.0, 1.0, 1.0)
					glTranslatef(-1.0 + chn, 0.5, 0.0)
					self.drawWave(dataRaw[chn][i:], l=NUM_SAMPLES)
					#self.drawWave(screen, dataRaw, yStart=0, height=height, col=(0xa0, 0xa0, 0xa0))
					glPopMatrix()
					"""
				
				glPushMatrix()
				glTranslatef(-1.0 + (chn * (1.0+SPACE/2)), -1.0, 0.0)
				if (CHANNELS > 1 and (not MONO)): glScalef(1.0 - SPACE/2, 1.0, 1.0)	# Skew a bit (gap in the middle)
				
				global SHOW_SPECTRUM_GRID
				if SHOW_SPECTRUM_GRID:
					
					# Generate display list since both sides are identical and don't change every frame!
					if gridList == None:
						gridList = glGenLists(1)
						glNewList(gridList, GL_COMPILE)
						
						glColor4f(*GRID_COLOR)
						
						glPushMatrix()
						glScalef(2.0/(CHANNELS if not MONO else 1), 1.0, 1.0)
						# Draw Volume Grid per channel
						#glBegin(GL_LINES)
						for i in xrange(SPECTRUM_GRID_STEPS_VOL):
							vol = 64.0 / (2.0**i)
							#y = volToY(vol) * height
							#glVertex2f(0.0, y)
							#glVertex2f(1.0, y)
							glBegin(GL_LINE_STRIP)
							for j in xrange(SPECTRUM_GRID_STEPS_FRQ):
								#band = int(NUM_BANDS / (2.0**(float(j)/2.0)))
								band = int(NUM_BANDS / (2.0**(float(j))))
								x = bandToXCache[band]	#bandToX(band)
								if (PHON_CORRECTION):	y = volToY(vol * phonCache[band]) * height
								else: y = volToY(vol) * height
								glVertex2f(x, y)
							glEnd()
						glPopMatrix()
						for i in xrange(SPECTRUM_GRID_STEPS_VOL):
							vol = 64.0 / (2.0**i)
							#y = volToY(vol) * height
							if (PHON_CORRECTION):	y = volToY(vol * phonCache[0]) * height
							else: y = volToY(vol) * height
							self.drawText(str(int(round( math.log(vol/64.0, 2)*10))) + " dB", (SPACE*0, y), rot=0)
						
						
						# Draw Frequency Grid per channel
						glBegin(GL_LINES)
						for i in xrange(SPECTRUM_GRID_STEPS_FRQ):
							#band = 2**i
							#band = int(NUM_BANDS / (2.0**(i/2.0)))
							band = int(NUM_BANDS / (2.0**(i)))
							x = width * bandToXCache[band]	#bandToX(band)
							glVertex2f(x, 0.0)
							glVertex2f(x, height)
						glEnd()
						for i in xrange(SPECTRUM_GRID_STEPS_FRQ):
							#band = 2**i
							#band = int(NUM_BANDS / (2.0**(i/2.0)))
							band = int(NUM_BANDS / (2.0**(i)))
							x = width * bandToXCache[band]	#bandToX(band)
							self.drawText(frqToStr( (SAMPLE_RATE * band) / (2*NUM_BANDS * TIME_ZOOM) ), (x, height), rot=45)
						glEndList()
					glCallList(gridList)
					
				glScalef(2.0 / (CHANNELS if not MONO else 1), height, 1.0)
				
				global SHOW_SPECTRUM, SHOW_SPECTRUM_COLOR
				if SHOW_SPECTRUM:
					# Draw current spectrum
					self.drawBars(spectrum[chn], avgData[chn], xCache=bandToXCache)
				
				global SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2
				if SHOW_LINES_MAX:	# Draw line: Peak
					self.drawLine(peakData[chn], xCache=bandToXCache, col=SPECTRUM_PEAK_COLOR)
				if SHOW_LINES_AVG:	# Draw line: Average
					self.drawLine(avgData[chn], xCache=bandToXCache, col=SPECTRUM_AVG_COLOR)
				if SHOW_LINES_AVG2:	# Draw line: Slow Average
					self.drawLine(avg2Data[chn], xCache=bandToXCache, col=SPECTRUM_AVG2_COLOR)
				glPopMatrix()
				
				
				# Stereo side-by-side
				glPushMatrix()
				glTranslatef(0.0, -0.33 + SPACE*2, 0.0)
				glRotatef(90.0, 0.0, 0.0, 1.0)
				glScalef(1.00, 0.25 * (1.0 - 2*chn), 0.25)	# Careful! this is rotated!
				global SHOW_SIDE_BY_SIDE
				if (SHOW_SIDE_BY_SIDE) and (CHANNELS > 1):	# and (not MONO):
					#self.drawBars(spectrum[chn], avgData[chn], xCache=bandToXCache)
					#self.drawBars(spectrum[chn], spectrum[chn], xCache=xCache)
					#self.drawBars(avgData[chn], spectrum[chn], xCache=bandToXCache)
					self.drawBars(spectrum[chn], spectrum[chn], xCache=bandToXCacheLin)
					
					glColor4f(*GRID_COLOR)
					glBegin(GL_LINES)
					glVertex2f(0.0, 0.0)
					glVertex2f(1.0, 0.0)
					glEnd()
					
				global SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2
				if SBS_SHOW_LINES_MAX:	# Draw line: Peak
					self.drawLine(peakData[chn], xCache=bandToXCacheLin, col=SPECTRUM_PEAK_COLOR)
				if SBS_SHOW_LINES_AVG:	# Draw line: Average
					self.drawLine(avgData[chn], xCache=bandToXCacheLin, col=SPECTRUM_AVG_COLOR)
				if SBS_SHOW_LINES_AVG2:	# Draw line: Slow Average
					self.drawLine(avg2Data[chn], xCache=bandToXCacheLin, col=SPECTRUM_AVG2_COLOR)
				
				if SBS_SHOW_LINES_DIFF and (chn == CHANNELS-1):
					self.drawLine(diffData, xCache=bandToXCacheLin, col=SPECTRUM_DIFF_COLOR)
				
				
				glPopMatrix()
			
			"""
			# Debug ring buffer
			#for chn in xrange(CHANNELS):
			# Debug
			glPushMatrix()
			glTranslatef(-0.5, 0.0, 0.0)
			glScalef(1.0, 0.2, 1.0)
			#self.drawWave([ringData[0][0] for k in xrange(NUM_SAMPLES)], NUM_SAMPLES)
			#self.drawWave(data[chn], NUM_SAMPLES2)
			
			glTranslatef(0.0, 2.0*chn, 0.0)
			glColor3f(0.0, 1.0, 0.0)
			self.drawWave(ringData[chn], RING_SIZE)
			#if (samples > 0): self.drawWave(dataRaw[chn], samples)
			glPopMatrix()
			"""
			
			# Superimpose Text
			
			#glColor4f(*INDICATOR_COLOR)
			#self.drawText(screen, lastPut, (SPACE, SPACE))
			
			global SHOW_FPS
			if SHOW_FPS and (tick > 0):
				glColor4f(*SIGNAL_COLOR)
				self.drawText(str(int(1000/tick)) + " fps", (-1.0 + SPACE, 1.0 - SPACE/2))	#	of max " + str(round(SAMPLE_RATE / NUM_SAMPLES )) + " fps")
			
			if (infoList == None):
				infoList = glGenLists(1)
				glNewList(infoList, GL_COMPILE)
				glColor4f(*SIGNAL_COLOR)
				#self.drawText(VERSION_TEXT, (-1.0 + SPACE, 1.0))
				
				#self.drawText(str(NUM_BANDS) + " bands, x" + str(TIME_ZOOM) + " zoom", (-1.0 + SPACE, 1.0 - SPACE))	#	of max " + str(round(SAMPLE_RATE / NUM_SAMPLES )) + " fps")
				#self.drawText(str(NUM_BANDS) + " bands, " + frqToStr(SAMPLE_RATE) + ", 1:" + str(TIME_ZOOM) + ", 0..." + frqToStr((SAMPLE_RATE/2)/TIME_ZOOM) + " - " + VERSION_TEXT, (-1.0, -1.0 - SPACE))
				#self.drawText(str(NUM_BANDS) + " bands, " + frqToStr(SAMPLE_RATE) + ":" + str(TIME_ZOOM) + " " + ("mono" if (MONO or CHANNELS < 2) else "stereo") + " < " + frqToStr((SAMPLE_RATE/2)/TIME_ZOOM), (-1.0, -1.0 - SPACE))
				self.drawText(frqToStr(SAMPLE_RATE) + ":" + str(TIME_ZOOM) + ", " + ("mono" if (MONO or CHANNELS < 2) else "stereo") + ", " + str(NUM_BANDS) + " bands ", (-1.0, -1.0 - SPACE))
				glEndList()
			global SHOW_INFO
			if SHOW_INFO:
				glCallList(infoList)
			
			global SHOW_HELP
			if (SHOW_HELP):
				glCallList(helpList)
			
			# Handle events
			#pressed = pygame.key.get_pressed()
			
			for event in pygame.event.get():
				if event.type == QUIT:
					exit()
				
				if event.type == VIDEORESIZE:
					#global SCREEN_SIZE
					put("Resize: " + str(event.size))
					SCREEN_SIZE = event.size
					glPopMatrix()
					self.resize(SCREEN_SIZE[0], SCREEN_SIZE[1])
					glPushMatrix()
					
				if event.type == MOUSEBUTTONDOWN:
					pass
				
				if event.type == MOUSEMOTION:
					#pygame.draw.line(screen, (0x33, 0xcc, 0x99), (oldMousePos[0], oldMousePos[1]), (event.pos[0], event.pos[1]), 2)
					#oldMousePos = event.pos
					#trace("x=" + str(event.pos[0]) + ", y=" + str(event.pos[1]), 0, False)
					pass
					
				if event.type == KEYDOWN:
					if (event.key == 27):
						running = False
					elif (event.key == 48):	# 0 faaaast!
						NUM_BANDS = 512
						TIME_ZOOM = 4
						MONO = True
						SONO = False
						PHASE = False
						SHOW_SIDE_BY_SIDE = False
						SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2 = False, False, False
						SBS_SHOW_LINES_DIFF = False
						SHOW_CHANNEL_WAVES = False
						SHOW_SPECTRUM_COLOR = False
						SHOW_SPECTRUM_GRID = False
						SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2 = False, False, False
						reCache = True
					elif (event.key == 49):	# 1 default!
						NUM_BANDS = 512
						TIME_ZOOM = 3
						MONO = False
						PHASE = True
						SONO = True
						SHOW_SIDE_BY_SIDE = True
						SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2 = False, True, False
						SBS_SHOW_LINES_DIFF = False
						SHOW_SPECTRUM_COLOR = True
						SHOW_SPECTRUM_GRID = True
						SHOW_CHANNEL_WAVES = True
						SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2 = True, True, True
						reCache = True
					elif (event.key == 50):	# 2 vocal!
						NUM_BANDS = 512	#1024
						TIME_ZOOM = 10
						MONO = True
						SONO = False	#True
						SHOW_SIDE_BY_SIDE = False
						SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2 = False, False, False
						SBS_SHOW_LINES_DIFF = False
						SHOW_SPECTRUM_COLOR = True	#False
						SHOW_SPECTRUM_GRID = True
						SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2 = True, True, False
						reCache = True
					elif (event.key == 51):	# 3 Nice
						NUM_BANDS = 128
						TIME_ZOOM = 3
						MONO = False
						PHASE = False
						SONO = True
						SHOW_SIDE_BY_SIDE = True
						SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2 = False, False, False
						SBS_SHOW_LINES_DIFF = True
						SHOW_SPECTRUM_COLOR = True
						SHOW_SPECTRUM_GRID = False
						SHOW_CHANNEL_WAVES = True
						SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2 = False, True, False
						reCache = True
					elif (event.key == 52):	# 4 HtKs favorite
						NUM_BANDS = 128
						TIME_ZOOM = 4
						MONO = False
						PHASE = True
						SONO = not True
						SHOW_SIDE_BY_SIDE = True
						SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2 = True, True, True
						SBS_SHOW_LINES_DIFF = False
						SHOW_SPECTRUM_COLOR = True
						SHOW_SPECTRUM_GRID = True
						SHOW_CHANNEL_WAVES = True
						SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2 = True, True, True
						reCache = True
					#elif (event.key == 53):	# 5
					elif (event.key == 54):	# 6
						SHOW_MASTER_WAVE = not SHOW_MASTER_WAVE
					#elif (event.key == 55):	# 7
					#elif (event.key == 56):	# 8
					#elif (event.key == 57):	# 9
					
					
					elif (event.key == K_a):	SHOW_SPECTRUM = not SHOW_SPECTRUM
					elif (event.key == K_b):	SHOW_SIDE_BY_SIDE = not SHOW_SIDE_BY_SIDE
					elif (event.key == K_c):	SHOW_SPECTRUM_COLOR = not SHOW_SPECTRUM_COLOR
					elif (event.key == K_d):	SBS_SHOW_LINES_DIFF = not SBS_SHOW_LINES_DIFF
					elif (event.key == K_e):	SHOW_ELIMINATION = not SHOW_ELIMINATION
					elif (event.key == K_f):	SHOW_FPS = not SHOW_FPS
					elif (event.key == K_g):	SHOW_SPECTRUM_GRID = not SHOW_SPECTRUM_GRID
					elif (event.key == K_l):
						# Get current lines state (depending on shift state)
						if (pygame.key.get_mods() & KMOD_SHIFT):
							v = (SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2)
						else:
							v = (SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2)
							
						if   (v == (True ,True ,True )): v = (False,False,False)
						elif (v == (False,False,False)): v = (False ,True ,False)
						elif (v == (False,True ,False)): v = (True ,True ,False)
						else: v = (True ,True ,True )
						
						if (pygame.key.get_mods() & KMOD_SHIFT):
							SBS_SHOW_LINES_MAX, SBS_SHOW_LINES_AVG, SBS_SHOW_LINES_AVG2 = v
						else:
							SHOW_LINES_MAX, SHOW_LINES_AVG, SHOW_LINES_AVG2 = v
					elif (event.key == K_h):	SHOW_HELP = not SHOW_HELP
					elif (event.key == K_i):	SHOW_INFO = not SHOW_INFO
					elif (event.key == K_m):
						MONO = not MONO if CHANNELS > 1 else True
						reCache = True
					elif (event.key == K_p):
						PHON_CORRECTION = not PHON_CORRECTION
						put("PHON_CORRECTION:" + str(PHON_CORRECTION))
						reCache = True
					elif (event.key == K_s):	SONO = not SONO
					elif (event.key == K_v):	PHASE = not PHASE
					elif (event.key == K_w):	SHOW_CHANNEL_WAVES = not SHOW_CHANNEL_WAVES
					elif (event.key == K_UP):	VIEW_POS[1] = VIEW_POS[1] + 0.01
					elif (event.key == K_DOWN):	VIEW_POS[1] = VIEW_POS[1] - 0.01
					elif (event.key == K_LEFT):	VIEW_POS[0] = VIEW_POS[0] - 0.01
					elif (event.key == K_RIGHT):	VIEW_POS[0] = VIEW_POS[0] + 0.01
					elif (event.key == K_KP_PLUS) or (event.key == K_PLUS):	VIEW_POS[2] = VIEW_POS[2] - 0.01
					elif (event.key == K_KP_MINUS) or (event.key == K_MINUS):	VIEW_POS[2] = VIEW_POS[2] + 0.01
					elif (event.key == K_KP_MULTIPLY) or (event.key == K_KP_DIVIDE):
						if (event.key == K_KP_MULTIPLY):
							TIME_ZOOM = max(TIME_ZOOM -1, 1)
						elif (event.key == K_KP_DIVIDE):
							TIME_ZOOM = min(TIME_ZOOM +1, MAX_TIME_ZOOM)
						reCache = True
					elif (event.key == K_PAGEUP) or (event.key == K_PAGEDOWN):
						if (event.key == K_PAGEUP):
							NUM_BANDS = min(MAX_BANDS, NUM_BANDS * 2)
						elif (event.key == K_PAGEDOWN):
							NUM_BANDS = max(2, NUM_BANDS / 2)
						reCache = True
					else:
						put(str(event.key))
					
				if event.type == KEYUP:
					pass
					
				# end of event loop
			
			#pygame.display.update()	# 2D
			pygame.display.flip()	# OpenGL
			glPopMatrix()
			
		put("Finished.")
		#sys.exit()


if __name__ == "__main__":
	try:
		pypFFT()
	finally:
		if streamIn is not None:
			streamIn.close()
